/**
 * @file controller_kp.c
 * @Author Kirill Ustimov <ustimov@cs.karelia.ru>
 * @date   Created on April 15, 2013
 * @brief  Publisher of pseudo-random generated values for sensors using http://www.w3.org/2005/Incubator/ssn/ssnx/ssn ontology
 *
 * This is file is a part of the C KPI library project and it has the same licence. 
 *
 * @section LICENSE
 * 
 * C KPI library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * C KPI Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * This file is a part of the 'Hello Sensors' example.
 *
 * Copyright (C) SmartSlog Team.
 * All rights reserved.
 * Mail-list: smartslog@cs.karelia.ru
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include <ckpi/ckpi.h>

// ---Constants---
// Describes connection to the SS.
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"
//#define KP_SS_ADDRESS "192.168.112.108"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "192.168.56.3"  // Local SIB
#define KP_SS_PORT 10010

// Name of the KP.
#define KP_NAME "HelloController"

// Parts of the triple for the inserting.
#define KP_NS_URI "http://X/Hello/#"
#define KP_SUBJECT KP_NS_URI"KP"
#define KP_PREDICATE KP_NS_URI"says"

// The word(object of the triple) for the exiting.
#define KP_EXIT_WORD "exit"

// After KP has publushed 'exit' keyword, it waits for a some time, 
// before cleaning the smart space. This time gives a chance to 
// KP-consumers, that a work without subscription receiving 'exit' keyword and 
// ends their work.
#define KP_WAITING_TIME_CLEANING 5   // For sleep() function

static bool keepRunning = true;

void sigHandler(int sig) {
    keepRunning = false;
}


/**
 * @brief Starts the KP, initializes state and cals all necessary functions.
 *
 * @return 0.
 */
int main()
{
    // Initialize smart space information.
    ss_info_t info;
    
    int nob;
    int i, n;
int res_value[300] = { 0 };
    ss_sparql_result_t * results;
    ss_sparql_result_t * o_results;
    ss_sparql_result_t * values_results;
    ss_sparql_result_t * orig_results;
    ss_sparql_result_t * orig_o_results;
    ss_sparql_result_t * orig_values_results;
    ss_subs_info_t sensors_sub_info;
    ss_subs_info_t values_sub_info[300];
    char sensor_uris[300][SS_URI_MAX_LEN];
    char sensor_measurements[300][SS_URI_MAX_LEN];
    int statuses[300] = { 0 };
int offset;
    char  hasValue_query[5000];
    char  sensed_data[6000];
   signal(SIGINT, sigHandler);
    ss_init_space_info(&info, KP_SS_NAME, KP_SS_ADDRESS, KP_SS_PORT);
   
    // In the smart space the name of the KP must be unique, if you try to 
    // join without leave, then you received an error.
    if (ss_join(&info, KP_NAME) == -1) {
        printf("Can't join to SS\n");
        return 0;
    }
    printf("KP join to SS. Press ctrl+C to exit.\n");
    // basic subscribe operation on sensors
    if (ss_sparql_select_subscribe(&info, &sensors_sub_info, "SELECT ?sensor_uri WHERE\
    {?sensor_uri <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://purl.oclc.org/NET/ssnx/ssn#Sensor>}", &results, &nob) == -1)
    {
      printf("error subscribe sensors\n");
      keepRunning = false;
    }
    n = 0;
	orig_results = results;
    while (results != NULL)
    {
	
	 
      //subscribe on each sensor publication
      sprintf(hasValue_query, "SELECT ?sensor_measurement WHERE {<%s> <http://purl.oclc.org/NET/ssnx/ssn#hasMeasurementProperty> ?sensor_measurement}", results->value[0]);
      ss_sparql_select_query(&info, hasValue_query, &values_results, &nob);
      if (values_results != NULL)
      {
	//printf("%s\n", values_results->value[0]);
	strcpy(sensor_measurements[n], values_results->value[0]);
      }
      ss_delete_sparql_results(values_results);
      values_results = NULL;
      sprintf(hasValue_query, "SELECT ?sensor_value WHERE {<%s> <http://purl.oclc.org/NET/ssnx/ssn#hasValue> ?sensor_value}", results->value[0]);
      if (ss_sparql_select_subscribe(&info, &values_sub_info[n], hasValue_query, &values_results, &nob) == -1)
      {
	printf("error subscribe values\n");
	keepRunning = false;
	break;
      }
      //printf("Sensor %s joined SS. Sensor hasMeasurmentProperty: %s, hasValue %s\n", results->value[0], sensor_measurements[n], (values_results != NULL) ? values_results->value[0] : "No value");
      strcpy(sensor_uris[n], results->value[0]);
      ss_delete_sparql_results(values_results);
      values_results = NULL;
      results = results->next;
      n++;
    }
    ss_delete_sparql_results(orig_results);
    orig_results = NULL;
    while (keepRunning)
    {	
      if (ss_sparql_select_subscribe_indication(&info, &sensors_sub_info, &results, &o_results, 1000) == 1)
      {
	orig_results = results;
	while (results != NULL)
	{
	  for (i = 0; i < n; i++)
	  {
	    if (strcmp(sensor_uris[i], results->value[0]) == 0)
	    {
	      results = results->next;
	      break;
	    }
	  }
	  if (i == n)
	  {
	    //subscribe on newely appeared sensors
	    sprintf(hasValue_query, "SELECT ?sensor_measurement WHERE {<%s> <http://purl.oclc.org/NET/ssnx/ssn#hasMeasurementProperty> ?senmasor_measurement}", results->value[0]);
	    	
	    ss_sparql_select_query(&info, hasValue_query, &values_results, &nob);
	    if (values_results != NULL)
	    {
	      strcpy(sensor_measurements[n], values_results->value[0]);
	    }
	    ss_delete_sparql_results(values_results);
	    values_results = NULL;
	    sprintf(hasValue_query, "SELECT ?sensor_value WHERE {<%s> <http://purl.oclc.org/NET/ssnx/ssn#hasValue> ?sensor_value}", results->value[0]);
	    if (ss_sparql_select_subscribe(&info, &values_sub_info[n], hasValue_query, &values_results, &nob) == -1)
	    {
	      printf("error subscribe new sensors\n");
	    keepRunning = false;
	    break;
	    }
	    printf("Sensor %s joined SS. Sensor hasMeasurmentProperty: %s, hasValue %s\n", results->value[0], sensor_measurements[n], (values_results != NULL) ? values_results->value[0] : "No value");
	    ss_delete_sparql_results(values_results);
	    strcpy(sensor_uris[n], results->value[0]);
	    values_results = NULL;
	    results = results->next;
	    n++;
	  }
	}
	ss_delete_sparql_results(orig_results);
	orig_results = NULL;
	/*while (o_results != NULL)
	{ 
	  for (i = 0; i < n; i++)
	  {
	    if (strcmp(sensor_uris[i], o_results->value[0]) == 0)
	    {
	      ss_unsubscribe(&info, &values_sub_info[i]);
	      statuses[i] = 1;
	      printf("Sensor %s has left SS\n", sensor_uris[i]);
	      break;
	    }
	  }
	  o_results = o_results->next;
	  }*/
	  ss_delete_sparql_results(o_results);
	  o_results = NULL;
	}
			offset = 0;
	strcpy(sensed_data, "");
	for (i = 0; i < n; i++)
	{
	  if (statuses[i] != 1)
	  {
	    if (ss_sparql_select_subscribe_indication(&info, &values_sub_info[i], &results, &o_results, 500) == 1)
	    {
		orig_results = results;

	      while (results != NULL)
	      {
		res_value[i] = atoi(results->value[0]);
		results = results->next;
	      }
		ss_delete_sparql_results(orig_results);

	      /*while (o_results != NULL)
	      {
		printf("Sensor %s has old value %s\n", sensor_uris[i], o_results->value[0]);
		o_results = o_results->next;
	      }*/
	      ss_delete_sparql_results(o_results);
	      o_results = NULL;
	    }		
		offset += sprintf(sensed_data + offset, "Sensor %s has measurement type %s, value %d\n", sensor_uris[i], sensor_measurements[i], res_value[i]);
	  }
	}
	if (offset != 0)
	{
		system("clear");

	printf("%s\n", sensed_data);
	}
	sleep(1);
    }
    for (i = 0; i < n; i++)
    {
      ss_unsubscribe(&info, &values_sub_info[i]);
    }
    ss_unsubscribe(&info, &sensors_sub_info);
    ss_leave(&info);
    printf("KP leave SS...\n");

    return 0;
}
