/**
 * @file consumer_kp.c
 * @Author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on August 3, 2012
 * @brief  Cunsumer of the example 'Hello World' without subscription.
 *
 * This is file is a part of the C KPI library project and it has the same licence. 
 *
 * @section LICENSE
 * 
 * C KPI library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * C KPI Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * This file is a part of the 'Hello World' example without subscription.
 * This KP makes aqueries to triples using <KP - says - *> triple-pattern.
 * It gets triples from the smart space and writes them. If the one of triples 
 * has an 'exit'  object, thenthis KP leaves the smart space and ends.
 * To stop the KP use CTRL+C 
 *
 * Copyright (C) SmartSlog Team (Aleksandr A. Lomov).
 * All rights reserved.
 * Mail-list: smartslog@cs.karelia.ru
 */

#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <ckpi/ckpi.h>

// ---Constants---
// Describes connection to the SS.
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "192.168.56.3"  // Local SIB
#define KP_SS_PORT 10010

// Name of the KP.
#define KP_NAME "HelloConsumer"

// Parts of the triple for the query.
#define KP_NS_URI "http://X/Hello/#"
#define KP_SUBJECT KP_NS_URI"KP"
#define KP_PREDICATE KP_NS_URI"says"

// The word(object of the triple) for the exiting.
#define KP_EXIT_WORD "exit"

// Time to sleep between cheking(query) a triple.
#define KP_WAITING_TIME 4   // For sleep() function


/****************** Static functions list **********************/
static void checking_triple(ss_info_t *info);
/*******************************************************/
        

/**
 * @brief Starts the KP, initializes state and cals all necessary functions.
 *
 * @return 0.
 */
int main()
{
    // Initialize smart space information.
    ss_info_t info;
    ss_init_space_info(&info, KP_SS_NAME, KP_SS_ADDRESS, KP_SS_PORT);
   
    // In the smart space the name of the KP must be unique, if you try to 
    // join without leave, then you received an error.
    if (ss_join(&info, KP_NAME) == -1) {
        printf("Can't join to SS\n");
        return 0;
    }

    printf("KP join to SS.\n");
    
    // Start to chek triples: create a triple for the query 
    // and make query op[eration to the smart space.
    checking_triple(&info);

    ss_leave(&info);    

    printf("\nKP leave SS...\n");

    return 0;
}

/**
 * @brief Checks triples in the SS.
 *
 * It queries triples and write information about them.
 * If it receives a KP_EXIT_WORD as value of the object in some triple, then 
 * it stops checking.
 *
 * @param[in] info information about the smart space..
 */
static void checking_triple(ss_info_t *info) 
{  
    if (info == NULL) {
        return;
    }
    
    ss_triple_t *req_triple = NULL;
    
    // Create the triple for a query operation.    
    ss_add_triple(&req_triple, KP_SUBJECT, KP_PREDICATE, SS_RDF_SIB_ANY, 
                SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
        
    if (req_triple == NULL) {
        printf("Not enough memory to create a triple.\n");
        return;
    }
    
    bool is_checking = true;
    
    while (is_checking == true) {
    
        ss_triple_t *result_triples = NULL;
    
        if (ss_query(info, req_triple, &result_triples) != 0) {
           printf("Can't perfome query\n");
           break;            
        }
    
        ss_triple_t *triple = result_triples;
        
        // Print information using triples that were receievng by the query.
        while (triple != NULL) {
            printf("KP says: '%s'.\n", triple->object);            
        
            // If there is the KP_EXIT_WORD as object, then stop checking.
            if (strcmp(KP_EXIT_WORD, triple->object) == 0) {
                printf("It's time to leave...\n");         
                is_checking = false;
                break;
            }        
        
            triple = triple->next;
        }
       
        ss_delete_triples(result_triples);
        
        // Sleep for a some time between cheking.
        sleep(KP_WAITING_TIME);       
    }
    
    ss_delete_triples(req_triple);
}
