/**
 * @file publisher_kp.c
 * @Author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on August 19, 2012
 * @brief  Publisher of the example 'Hello World' with subscription.
 *
 * This is file is a part of the C KPI library project and it has the same licence. 
 *
 * @section LICENSE
 * 
 * C KPI library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * C KPI Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * This file is a part of the 'Hello World' example with subscription.
 * This KP updates a triple <KP - says - text>, where 'text' is a input from 
 * the user. 
 *
 * Copyright (C) SmartSlog Team (Aleksandr A. Lomov).
 * All rights reserved.
 * Mail-list: smartslog@cs.karelia.ru
 */
 
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <ckpi/ckpi.h>

// ---Constants---
// Describes connection to the SS.
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "192.168.5.85"  // Local SIB
#define KP_SS_PORT 10010

// Name of the KP.
#define KP_NAME "HelloPublisherSub"

// Parts of the triple for the inserting.
#define KP_NS_URI "http://X/Hello/#"
#define KP_SUBJECT KP_NS_URI"KP"
#define KP_PREDICATE KP_NS_URI"says"

// The word(object of the triple) for the exiting.
#define KP_EXIT_WORD "exit"


/****************** Static functions list **********************/
static void process_publish(ss_info_t *info);
static void remove_newline(char *str);         
/*******************************************************/


/**
 * @brief Starts the KP, initializes state and cals all necessary functions.
 *
 * @return 0.
 */
int main()
{
    ss_info_t info;
    
    // Initialize smart space information and data about subscription.
    ss_init_space_info(&info, KP_SS_NAME, KP_SS_ADDRESS, KP_SS_PORT);
   
   // In the smart space the name of the KP must be unique, if you try to 
   // join without leave, then you received an error.
    if (ss_join(&info, KP_NAME) == -1) {
        printf("Can't join to SS\n");
        return 0;
    }

    printf("KP join to SS.\n");
    
    // KP starts to update triples using object value that was input by user, 
    // this function works while KP_EXIT_WORD is not input.
    process_publish(&info);    

    ss_leave(&info);    

    printf("\nKP leave SS...\n");

    return 0;
}

/**
 * @brief Publish information to the smart space.
 *
 * Creates triples for removing and adding. 
 * The object of the inserting triple is reading from the sdin.
 * If the user input 'exit' keyword then it cleaning the smart space 
 * and ends working.
 *
 * @param[in] info information about the smart space..
 */
static void process_publish(ss_info_t *info) 
{  
    if (info == NULL) {
        return;
    }
    
    ss_triple_t *rem_triple = NULL;
        
    // Create triple for removing data, this triple is not changing while 
    // publishing process working.    
    ss_add_triple(&rem_triple, KP_SUBJECT, KP_PREDICATE, SS_RDF_SIB_ANY, 
                SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
        
    if (rem_triple == NULL) {
        printf("Not enough memory to create a triple.\n");
        ss_delete_triples(rem_triple);
        return;
    }

    // Read a string from the stdin and create a triple using this line as 
    // a value for the object. 
    // Update triple in the SS. 
    // If object equals KP_EXIT_WORD constant, then it waiting for a some time
    // and cleaning the SS.
    while (true) {
        char object_text[SS_OBJECT_MAX_LEN];
        ss_triple_t *ins_triple = NULL;

        printf("\nSay something or write '%s':\n", KP_EXIT_WORD);

        if (fgets(object_text, SS_OBJECT_MAX_LEN, stdin) == NULL) {
            printf("Somethign wrong with the input!\n");
            break;
        }    
        
        // Remove the '\n' character.
        remove_newline(object_text);
          
        // Create triple for inserting.
        ss_add_triple(&ins_triple, KP_SUBJECT, KP_PREDICATE, object_text, 
                SS_RDF_TYPE_URI, SS_RDF_TYPE_LIT);
 
        if (ins_triple == NULL) {
            printf("Not enough memory to create a triple.\n");
            break;
        }
                
        printf("Publishing: %s\n", ins_triple->object);
          
        int result = ss_update(info, ins_triple, rem_triple, NULL);
        
        ss_delete_triples(ins_triple);
               
        if (result != 0) {
            printf("Can't update triples.\n");
            break;
        } 
        
        // If the user typed KP_EXIT_WORD, then cleaning the SS.
        if (strcmp(KP_EXIT_WORD, object_text) == 0) {
            ss_remove(info, rem_triple);
            break;
        }               
    }
    
    ss_delete_triples(rem_triple);
}

/**
 *@brief Finds '\n' character in the end of the line and chenges it to '\0';
 * 
 * @param[in] str string to remove endline character.
 */
static void remove_newline(char *str) 
{
    if (str == NULL) {
        return;
    }
    
    size_t len = strlen(str);
    
    if (str[len -1] == '\n') {
        str[len -1] = '\0';
    }
}

